/* For proper formating set tabstop=3 */

#include <lpfgall.h>

/* Definitions of constants defining apex states */
#define MONOPODIAL 0
#define SYMPODIAL  1
#define FLOWERING  2

/* Wrappers for time functions defined graphically via timeline editor */

#define MERISTEM_SCALE			1.0*pleiotropic_scale
#define meristem_size(t)		MERISTEM_SCALE*tfunc(MERISTEM_DIAMETER, t)

#define INTERNODE_SCALE 		1.0*pleiotropic_scale
#define internode_length(t)     INTERNODE_SCALE*tfunc(INTERNODE_LENGTH, t)

#define BRACT_SCALE				1.0*pleiotropic_scale
#define bract_size(t)			BRACT_SCALE*tfunc(BRACT_SIZE, t)

#define PETIOLE_SCALE			1.0*pleiotropic_scale
#define petiole_width(t)		PETIOLE_SCALE*tfunc(PETIOLE_WIDTH, t)

#define FLOWER_SCALE			0.08*pleiotropic_scale
#define flower_size(t)			FLOWER_SCALE*tfunc(FLOWER_DIAMETER, t)

#define COLOR_SCALE				32
#define flower_color(t)			COLOR_SCALE*tfunc(FLOWER_COLOR, t)

/* Declarations of data structures, modules, and global variables */
struct MeristemData {
	int state;	// MONOPODIAL, SYMPODIAL, or FLOWERING
	float age;	// age since creation or production of the most recent primordium
	float d;	// diameter
};
module Meristem(MeristemData);

struct InternodeData {
	float age;		// age since the time of creation
	float v;		// veg at time of creation
	float dn;		// diameter raised to power n (pipe_model_power)
};
module Internode(InternodeData);

struct BractData {
	float age;		// age from the time of creation
	float v;		// veg at time of creation
	float d;		// diameter
};
module Bract(BractData);

module INIT();		// initializes variables used in the definition of initial structrue
module Primordium(float,float); // (age, veg difference between sibling primordia)
module StartPlant(float,float);

float max_age;				// inflorescence age at the end of simulation
float dt;					// simulation time step

float plastochron; 		// plastochron
float delay_mono;			// delay in the development of lateral primordia - raceme
float delay_sym;			// delay in the development of lateral primordia - cyme 
float delay_diff;			// difference in delay between cyme branches

float veg;					// vegetativeness 
float veg_init; 			// initial veg value (usually 1)
float veg_half_life;		// time of veg decline by to 1/2 ofveg_init 
float veg_decline_rate;	// computed from veg_half_life 

float th_m;					// threshold veg for the monopodial meristem switch to flowering
float th_s;					// threshold veg for arresting sympodial meristem production
float th_diff;				// veg difference vetween sybling primordia 

float b_th_1;				// bract threshold for switch from 3 lobes to 1 lobe
float b_th_2;				// bract threshold for switch from 1 lobe to narrow form

float pleiotropic_scale;		// controls overall inflorewcence size
float peduncle_length_ctrl;	// controls peduncle length by adjusting its veg
float int_len_power;		// controls internode length as a power function of veg
float internode_segments;		// number of segments subdividing an internode

float distal_diameter;	// meristem diameter for pipe model
float pipe_model_power;	// power for propagating internode diameters in the pipe model

float angle_main;			// deflection of the apex direction in the primary axis
float angle_lateral;		// branching angle - monopodial
float angle_sympodial;	// branching angle - sympodial

float int_elasticity; 			// internode "elasticity" 
float int_elasticity_power;	// controls internode elasticity as a power of veg

float bract_angle_branch;		// orients bract supporting lateral branch
float bract_angle_no_branch;	// orients bract not supporting a lateral branch

float ang_adjustment_1;	// branching angle adjustment for petiole supporting 3-lobed bract
float ang_adjustment_2;	// branching angle adjustment for petiole supportinf 1-lobe bract
float petiole_length_1;	// target petiole length for bracts with 3 lobes
float petiole_length_2;	// targetpetiole length for bracts with one love 
float petiole_elasticity_1; 	// petiole elasticity for bracts with 3 loges
float petiole_elasticity_2;	// petiole Valishirinelasticity for bracts with 1 lobe

float bract_size_power;	// controls the size of small bracts as a power function of veg

float start_age; // the minimum choronological age of first plant shown in calibration
float scale_factor_age; // scale factor mapping time in model to time shown in calibration

Start: 
{ 
	/* Input parameter values from the parameters.vset file */

	// simulation timing
	max_age = val(MAX_AGE);
	dt = val(DT);

	// branch initiation	
	plastochron = val(PLASTOCHRON);
	delay_mono = val(LATERAL_DELAY_MONO);
	delay_sym = val(LATERAL_DELAY_SYM);
	delay_diff = val(LATERAL_DELAY_DIFF);

	// veg control
	veg_init = val(VEG_INIT);
	veg = veg_init;
	veg_half_life = val(VEG_HALF_LIFE);
	veg_decline_rate = log(2.) / veg_half_life;

	// branching thresholds
	th_m = val(TH_MONO);
	th_s = val(TH_SYM);
	th_diff = val(TH_DIFF);

	// bract type thresholds
	b_th_1 = val(BRACT_TH_1);
	b_th_2 = val(BRACT_TH_2);

	// dimensions
	pleiotropic_scale = val(PLEIOTROPIC_SCALE);
	peduncle_length_ctrl = val(PEDUNCLE_LENGTH_CTRL);
    int_len_power = val(INT_LEN_POWER);
	internode_segments = val(INTERNODE_SEGMENTS);

	// pipe model
	distal_diameter = val(DISTAL_DIAMETER);
	pipe_model_power = val(PIPE_MODEL_POWER);

	// branching angles
	angle_main = val(ANGLE_MAIN);
	angle_lateral = val(ANGLE_LATERAL);
	angle_sympodial = val(ANGLE_SYMPODIAL);

	// gravity and tropism
	int_elasticity = val(INT_ELASTICITY);
	int_elasticity_power = val(INT_ELASTICITY_POWER);

	// bracts
	bract_angle_branch = val(BRACT_ANGLE_BRANCH);
	bract_angle_no_branch = val(BRACT_ANGLE_NO_BRANCH);

	ang_adjustment_1 = val(ANG_ADJUSTMENT_1);
	ang_adjustment_2 = val(ANG_ADJUSTMENT_2);
	petiole_length_1 = val(PETIOLE_LENGTH_1);
	petiole_length_2 = val(PETIOLE_LENGTH_2);
	petiole_elasticity_1 = val(PETIOLE_ELASTICITY_1);
	petiole_elasticity_2 = val(PETIOLE_ELASTICITY_2);

    bract_size_power = val(BRACT_SIZE_POWER);

    start_age = val(START_AGE);
    scale_factor_age = val(SCALED_AGE);

	/* Scan string backward as needed for the pipe model */
	Backward(); 
}

EndEach: 
{
    // the global variable veg is not used in this model because
    // a veg variable is assigned to each plant in the calibration
	veg -= veg*veg_decline_rate*dt;  
}

consider: Meristem Internode Bract ;	// modules considered in pipe model information flow

derivation length: (int) (max_age / dt);

Axiom: INIT();

/* Define, initialize and produce the initial structure */ 
INIT():
{
    const int NUM_PLANTS = 7;
    static char str[NUM_PLANTS][32];

    // draw background image
	const float SF = 0.0025; 
	float w = 2471.0 * SF; 
	float h = 644.0 * SF; 
    nproduce SB f(1.625) 
                Right(90) f(w*0.5) Left(90)
                SetColor(255) CurrentContour(1) CurrentTexture(4) TextureVCoeff(1.0/h)
                SetWidth(w) StartGC F(h) EndGC
             EB;

    // generate N plants with different intial ages
    for (int i = 0; i < NUM_PLANTS; i++) {

        // init a plant
        InternodeData InitInternodeData;
	    InitInternodeData.age = 0;
	    InitInternodeData.v = veg_init*peduncle_length_ctrl;
	    InitInternodeData.dn = 0;

	    MeristemData InitMeristemData;
	    InitMeristemData.state = MONOPODIAL;
	    InitMeristemData.age = 0;
	    InitMeristemData.d = 0;

        // the initial choronological age depends on the calibration images
        // it can be negative, to show younger plants
        nproduce SB;
        switch (i) {
            case 0:
                nproduce Right(90) f(0.2) Left(90) RollL(180);
                InitMeristemData.age = 0./scale_factor_age-max_age + start_age;
                break;
            case 1:
                nproduce Right(90) f(0.825) Left(90) RollL(180);
                InitMeristemData.age = 4./scale_factor_age-max_age + start_age;
                break;
            case 2:
                nproduce Right(90) f(1.675) Left(90) RollL(180);
                InitMeristemData.age = 7./scale_factor_age-max_age + start_age;
                break;
            case 3:
                nproduce Right(90) f(2.65) Left(90) RollL(180);
                InitMeristemData.age = 10./scale_factor_age-max_age + start_age;
                break;
            case 4:
                nproduce Right(90) f(3.55) Left(90) RollL(180);
                InitMeristemData.age = 12./scale_factor_age-max_age + start_age;
                break;
            case 5:
                nproduce Right(90) f(4.55) Left(90) RollL(180);
                InitMeristemData.age = 17./scale_factor_age-max_age + start_age;
                break;
            case 6:
                nproduce Right(90) f(5.7) Left(90) RollL(180);
                InitMeristemData.age = 23./scale_factor_age-max_age + start_age;
                break;
        }
        InitMeristemData.age = InitMeristemData.age > 1. ? 0.99 : InitMeristemData.age;
        InitInternodeData.age = InitMeristemData.age;
        nproduce SB StartPlant(-1.,InitMeristemData.age) 
                    Internode(InitInternodeData) Meristem(InitMeristemData)
                 EB;
        nproduce EB;
    }

    produce;
}

/* Decrease Veg for each plant in the calibration when its chronological age reaches 0 */
StartPlant(plantVeg,plantAge): {
    if (plantAge >= 0.) {
        plantVeg = plantVeg == -1. ? veg_init : plantVeg*(1. - veg_decline_rate*dt);
    }
    produce StartPlant(plantVeg,plantAge+dt);
}

/* Advance meristem age by time step dt, update meristem diameter */
Meristem(m):
{
	m.age += dt; 
 	m.d = fmax(m.d, distal_diameter * meristem_size(m.age));
	produce Meristem(m);
}

/* Advance bract age by time step dt, update petiole diameter */
Bract(b):
{
	b.age += dt; 
	b.d = distal_diameter * petiole_width(b.age);
	produce Bract(b);
}

/* Advance internode age by time step dt, update diameter according to the pipe model */
Internode(s): 
{
	s.age += dt;
	s.dn = 0;

	struct MeristemData mr;		// possible contexts : meristem...
	struct InternodeData sr;	// ...an internode
	struct BractData br;			// ...or a bract

	while (1) { // Sum up all contributions to the width of the supporting internode

		if (InNewRightContext(SB SB Bract(br) EB Internode(sr) EB)) {	// bract and supported branch
			s.dn += pow(br.d,pipe_model_power) + sr.dn;
			continue;
		}

		if (InNewRightContext(SB SB Bract(br) EB EB)) {	// bract not supporting a branch
			s.dn = pow(br.d,pipe_model_power);
			continue;
		}

		if (InNewRightContext(Meristem(mr))) {		// meristem
 			s.dn = pow(mr.d,pipe_model_power);
			break;
		}
		if (InNewRightContext(Internode(sr))) {	// child internode in current axis
			s.dn += sr.dn;
			break;
		}
	}

	produce Internode(s);
}

/* Main part of the model - see the object description */
decomposition:
maximum depth: 5;

consider: StartPlant;

/* Veg is not a global parameter. It is stored in the StartPlant module, */
/* which marks the start of each plant in the calibration. */
StartPlant(plantVeg,plantAge) < Meristem(m):
{
	if (m.age >= plastochron) {

		if (m.state == MONOPODIAL) {	// 1st order meristem: monopodial branching
			if (plantVeg > th_m) {				// high veg: produce a lateral meristem 
				m.age -= plastochron;
				InternodeData s = {m.age, plantVeg ,0};
				nproduce SB Left(angle_lateral) Primordium(m.age-delay_mono, plantVeg-th_s) EB;
				produce Right(angle_main) RollL(180) Internode(s) Meristem(m);
			}
			else {
				m.state = FLOWERING;		// low veg: change state to flowering 
				produce Meristem(m);
			}
		}

		if (m.state == SYMPODIAL)  {	// meristem of order 2 or higher: sympodial branching
			m.age -= plastochron;
			InternodeData s = {m.age, plantVeg, 0};
			nproduce SB Left(angle_sympodial) Primordium(m.age-delay_sym, plantVeg-th_s) EB;
			nproduce SB RollL(180) Left(angle_sympodial) 
				Primordium(m.age-delay_sym-delay_diff, plantVeg-th_s-th_diff) EB;
			m.state = FLOWERING;			// parent meristem always switches to flowering
			produce Internode(s) Meristem(m);
		}
	}
}
		
StartPlant(plantVeg,plantAge) < Primordium(p_age,p_veg) :
{
	BractData b = {p_age, plantVeg, (float) (bract_size(b.age))};
	if (p_veg > 0) {									// If veg is high enough
		MeristemData new_m = {SYMPODIAL, p_age};
		InternodeData s = {new_m.age, plantVeg, 0};
		nproduce SB Left(bract_angle_branch) Bract(b) EB; 	// produce a bract
		produce Internode(s) Meristem(new_m);	// and a meristem supported by an internode
	}
	else {												// If veg is not high enough
		nproduce SB Right(bract_angle_no_branch) Bract(b) EB;	// produce a bract 
		produce;											// and abort
	}
}
		
interpretation:
maximum depth: 15;

/* Declare modules used to visualize reproductive organs */
module Flower(float);
module Petal(float,float,float);
module Sepal(float,float,float);
module Episepal(float,float,float);
module Receptacle(float,float);
module Leaflet(float,float,float,int);

#define bract_unfolding(t) tfunc(BRACT_UNFOLDING,t)

/* Display the scaled choronological age of a plant corresponding to its calibration image */
StartPlant(plantVigor,plantAge) :
{
    static char scaled_age[32];
    sprintf(scaled_age,"%g", std::abs(std::round((plantAge - start_age)*scale_factor_age)));
    float x = -0.025;
    if ((plantAge - start_age)*scale_factor_age > 9)
        x -= 0.025;
    nproduce SB f(-0.125) Left(90) f(x) Right(90) SetColor(255) Label(scaled_age) EB;
    produce;
}

/* Visualize meristem in flowering state as a sphere with radius and color depending on age */
Meristem(m) :
{  
	if (m.state == FLOWERING) {
        if (val(SHOW_FLOWERS) == 0) {
		    float t = m.age;
		    produce SetColor(16+floor(flower_color(t)))
			        f(flower_size(t)) Sphere(flower_size(t));
        }
        else {
            produce SB SetCoordinateSystem(1) Flower(m.age) EB;
        }
	}
}

/* Visualize internode */
Internode(s):
{
	// Internode length is a function of age and vigor at the time of creation
	float int_length_veg = internode_length(s.age)*pow(s.v,int_len_power);
	// Internode width is set following the pipe model
	float d = pow(s.dn, 1.0/pipe_model_power);
	nproduce SetWidth(d);
	// Introduce some curvature to internodes, inspired by the gravity / tropisms.
	// The magnitude of this effect depends on vigor (purely for visual purposes).
	nproduce SetElasticity (0,(int_elasticity / pow(s.v, int_elasticity_power)) / internode_segments);
	// To capture this curvature, each internode is divided into segments.
	for (int i = 0; i<internode_segments; i++)
		nproduce F(int_length_veg/internode_segments);
	produce;
}

/* Visualize bract in a manner depending on veg */
Bract(b):
{
    // don't draw bracts that are too small to see
    if (bract_size(b.age) <= 1e-4)
        produce;

    float scale = val(GC_BRACT_SCALE);

    // set color
    nproduce SetColor(2);

	// If veg is high, create a leaf-like bract with three lobes
	if (b.v > b_th_1) { 
		nproduce Left(ang_adjustment_1-val(GC_BRACT_ANGLE_ADJ)+val(GC_BRACT_ANGLE_AGE)*bract_size(b.age));
		nproduce SetWidth (b.d);
		nproduce SetElasticity(0,petiole_elasticity_1 / internode_segments);
		float petiole_length = petiole_length_1 * bract_size(b.age);
		for (int i = 0; i<internode_segments; i++)
			nproduce F(petiole_length/internode_segments);
		//produce CurrentTexture(0) Surface(0,bract_size(b.age));
        produce RollR(180)
                SB 
                   Right(val(GC_LEAFLET1_ANG)*bract_unfolding(b.age)) 
                   RollL(val(GC_LEAFLET1_ROLL)*bract_unfolding(b.age))
                   Leaflet(b.age,scale*0.28*b.v,scale*0.28*b.v,0) 
                EB
                SB
                   Right(val(GC_LEAFLET2_ANG)*bract_unfolding(b.age)) 
                   RollL(val(GC_LEAFLET2_ROLL)*bract_unfolding(b.age)) 
                   Leaflet(b.age,scale*0.22*b.v,scale*0.25*b.v,1) 
                EB
                SB
                   Right(val(GC_LEAFLET3_ANG)*bract_unfolding(b.age)) 
                   RollL(val(GC_LEAFLET3_ROLL)*bract_unfolding(b.age)) 
                   Leaflet(b.age,scale*0.2*b.v,scale*0.25*b.v,2) 
                EB;
	}		

	// If veg is somewhat lower, create a bract with a single lobe
	if (b.v > b_th_2) { 
		nproduce Left(ang_adjustment_2);
		nproduce SetWidth (b.d);
		nproduce SetElasticity(0,petiole_elasticity_2 /internode_segments);
		float petiole_length = petiole_length_2 * bract_size(b.age);
		for (int i = 0; i<internode_segments; i++)
			nproduce F(petiole_length/internode_segments);
		//produce CurrentTexture(1) Surface(1,bract_size(b.age));
        produce RollR(180.f) Leaflet(b.age,scale*0.2*b.v,scale*0.3*b.v,0);
	}		

	// If veg is even less, create a bract with size depending on veg
	float bract_size_veg = bract_size(b.age)*pow(b.v,bract_size_power);
	produce RollL(90) Surface(0, bract_size_veg);
}

Leaflet(age,lenScale,widScale,texId) : {
        float len = lenScale*bract_size(age);
		produce BlendedContour(folded_contour,leaf_contour,bract_unfolding(age)) 
                CurrentTexture(texId) TextureVCoeff(1./(lenScale*bract_size(age)))
                SetWidth(widScale*bract_size(age)) 
                StartGC F(len*0.5) F(len*0.5) EndGC;
}


// flower/fruit drawing functions
#define episepal_growth(t) tfunc(EPISEPAL_GROWTH,t)
#define episepal_angle(t) 90.0*tfunc(EPISEPAL_ANGLE,t)
#define episepal_width(x) func(EPISEPAL_WIDTH,x)
#define episepal_bend(x) 90.0*func(EPISEPAL_BEND,x)

#define sepal_length_growth(t) tfunc(SEPAL_LENGTH_GROWTH,t)
#define sepal_width_growth(t) tfunc(SEPAL_WIDTH_GROWTH,t)
#define sepal_bend_time(t) tfunc(SEPAL_BEND_TIME,t)
#define sepal_angle(t) 90.0*tfunc(SEPAL_ANGLE,t)
#define sepal_width(x) func(SEPAL_WIDTH,x)
#define sepal_bend(x) 90.0*func(SEPAL_BEND,x)

#define petal_length_growth(t) tfunc(PETAL_LENGTH_GROWTH,t)
#define petal_width_growth(t) tfunc(PETAL_WIDTH_GROWTH,t)
#define petal_bend_time(t) tfunc(PETAL_BEND_TIME,t)
#define petal_angle(t) 90.0*tfunc(PETAL_ANGLE,t)
#define petal_width(x) func(PETAL_WIDTH,x)
#define petal_bend(x) 90.0*func(PETAL_BEND,x)

#define fruit_growth(t) tfunc(FRUIT_GROWTH,t)
#define fruit_color(t) tfunc(FRUIT_COLOR,t)

Flower(age):
{
  float scale = val(GC_FLOWER_SCALE) * val(PLEIOTROPIC_SCALE);
  // twist the flowers towards the camera...
  nproduce SetElasticity(0,0.) SetElasticity(1,0.3) F(0.0001) F(0.0001) F(0.0001) SetElasticity(1,0.);

  // epicalyx - five green episepals (group of bracts resembling a calyx)
  nproduce SB CurrentContour(episepal_contour) SetColor(4);
  for (int i = 0; i < 5; i++) 
    nproduce SB RollL(72*i) Episepal(age,val(GC_LEN_SCALE)*scale,val(GC_WID_SCALE)*scale) EB;
  nproduce EB;

  // calyx - five green sepals
  nproduce SB CurrentContour(sepal_contour) RollL(72*0.5) SetColor(4);
  for (int i = 0; i < 5; i++) 
    nproduce SB RollL(72*i) Sepal(age,val(GC_LEN_SCALE)*scale,val(GC_WID_SCALE)*scale) EB;
  nproduce EB;

  // corolla - five white petals
  nproduce f(scale*0.1);
  nproduce SB CurrentContour(petal_contour) SetColor(3);
  for (int i = 0; i < 5; i++) 
    nproduce SB RollL(72*i) Petal(age,val(GC_PETAL_LEN_SCALE)*scale,scale) EB;
  nproduce EB;

  // receptacle (fruit)
  nproduce f(scale*0.01);
  nproduce SB Down(90) Receptacle(age,val(GC_FRUIT_SCALE)*val(PLEIOTROPIC_SCALE)) EB;

  produce;
}

const float DX = 0.02f;

Episepal(age,len,wid): {
    // set opening anlge
    nproduce Down(episepal_angle(age));

    len = len * episepal_growth(age);
    wid = wid * episepal_growth(age);

    // set the initial petal width
    nproduce SetWidth(wid*episepal_width(0));
    // start drawing the generalized cylinder
	nproduce  
		StartGC();
        for (float x = 0.; x < 1.; x += DX) { 
            // generate leaf segment
            nproduce F(len*DX); 
            // set width and orientation of next segment
		    nproduce SetWidth(wid*episepal_width(x)) 
                    Up(episepal_bend(x)*DX) 
                    ; 
        } 
    produce EndGC(); // end the generalized cylinder
}

Sepal(age,len,wid): {
    // set opening anlge
    nproduce Down(sepal_angle(age));

    len = len * sepal_length_growth(age);
    wid = wid * sepal_width_growth(age);

    // set the initial petal width
    nproduce SetWidth(wid*sepal_width(0));
    // start drawing the generalized cylinder
	nproduce  
		StartGC();
        for (float x = 0.; x < 1.; x += DX) { 
            // generate leaf segment
            nproduce F(len*DX); 
            // set width and orientation of next segment
		    nproduce SetWidth(wid*sepal_width(x)) 
                    Up(sepal_bend_time(age)*sepal_bend(x)*DX) 
                    ; 
        } 
    produce EndGC(); // end the generalized cylinder
}

Petal(age,len,wid): {
    // set opening anlge
    nproduce Down(petal_angle(age));

    // drop petal after certain age
    if (age >= val(MAX_PETAL_AGE)) 
        produce;

    len = len * petal_length_growth(age);
    wid = wid * petal_width_growth(age);
    
    // set the initial petal width
    nproduce SetWidth(wid*petal_width(0));
    // start drawing the generalized cylinder
	nproduce  
		StartGC();
        for (float x = 0.; x < 1.; x += DX) { 
            // generate leaf segment
            nproduce F(len*DX); 
            // set width and orientation of next segment
		    nproduce SetWidth(wid*petal_width(x)) 
                    Up(petal_bend_time(age)*petal_bend(x)*DX) 
                    ; 
        } 
    produce EndGC(); // end the generalized cylinder
}

Receptacle(age,wid): {
  // draws longitudinal sections of a surface of revolution

  const int texture_repeats = 4; // number of times the texture repeats on the surface
  const int transversal_divisions = 40; // set in ContourSides
  const float longitudinal_divisions = 40.f; // sets longitudinal sections
  float ang_incr = 360.f / longitudinal_divisions; 

  // setup the contour and texturing
  nproduce SetColor(15 + int(32.*fruit_color(age))) SetWidth(wid*fruit_growth(age));
  nproduce ContourSides(transversal_divisions);
  nproduce BlendedContour(receptacle_contour_one,receptacle_contour_two,std::min(std::max(0.f,fruit_growth(age)),1.f));
  nproduce CurrentTexture(3) TextureVCoeff(float(texture_repeats)/longitudinal_divisions);

  // draw the generalized cylinder as a surface of revolution
  nproduce SB StartGC;
  for (float ang = 0.f; ang < 360.f; ang += ang_incr) {
    nproduce Right(ang_incr) F(0);
  }
  nproduce EndGC EB;
  
  produce;
}
